from jnpr.junos import Device
from jnpr.junos.exception import ConnectError
from jnpr.junos.exception import LockError
from jnpr.junos.exception import UnlockError
from jnpr.junos.exception import CommitError
from jnpr.junos.exception import RpcError


def open_session(dev_list):
    try:
        dev = Device(host=dev_list[0], user=dev_list[1], password=dev_list[2]) # start Net conf
        print (">Opening session in progress....")
        dev.open() # open connention
        print ("Session opened succesfully.\n")
        return dev
    except ConnectError as err: 
        print ("Cannot connect to device: {0}".format(err))
        print(">Press any key to retry...")
        input()
        return 'retry'
            
def lock_configuration(cu,dev_list):
    print (">Locking the configuration in progress....")
    try:
        cu.lock()
        print ("Configuration locked succesfully.\n")
    except LockError as err:
        print ("Unable to lock configuration: {0}".format(err))
        print("Press any key to retry...")
        input()
        open_session(dev_list)
        
def commit_check(cu):
    try:
        #Commit check the loaded configuration
        print (">Commit check the loaded configuration in progress......")
        cu.commit_check(timeout=60)
        print ("Commit check ended succesfully with no errors.\n")
        print("Press any key to commit...")
        input()
    except CommitError as err:
        # try rollback
        try:
            print ("Error: Commit check the loaded configuration: {0}".format(err))
            print("Press any key to rollback...")
            input()
            print (">Rolling back the configuration [rollback 0]....")
            cu.rollback(rb_id=0)
            print ("Committing the configuration after rollback 0")
            cu.commit()
            print ("Committing the configuration after rollback 0 succesfully.")
            return
        except RpcError as err:
            print ("Unable to roll back configuration changes: {0}".format(err))
            print("Press any key to continue...")
            input()
            return
        except CommitError as err:
            print ("Error: Unable to commit configuration: {0}".format(err))
            print("Press any key to continue...")
            input()
            return


def commit(cu,Configuration_Description = ''):
    try:
        #Committing the configuration......
        print (">Committing the configuration......")
        cu.commit(comment=Configuration_Description, timeout=300)
        print ("Committing the configuration succesfully.\n")
    except CommitError as err:
        print ("Unable to commit configuration: {0}".format(err))
        
        
def unlock_close(cu, dev):
    try:
        print (">Unlocking the configuration....")
        cu.unlock()
        print ("Configuration Unlocked.\n")
    except UnlockError as err:
        print ("Unable to unlock configuration: {0}".format(err))

def unlock(cu):
    try:
        print (">Unlocking the configuration....")
        cu.unlock()
        print ("Configuration Unlocked.\n")
    except UnlockError as err:
        print ("Unable to unlock configuration: {0}".format(err))
  
def session_close(dev):
    # End the NETCONF session and close the connection
    dev.close()
    print ("Session Closed.")


from srx_classes import devices
from srx_classes import excel_output
from lxml import etree
import sys
import re

def policy_def(type_ = None , file_content=None ,dev=None,file_name=None, onlyData=None):
    returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
    data_to_return = returned[0]
    columns_ = returned[1]
    rules = save_policy(data_to_return,type_)
    sheet_name_ = 'Policies'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    } 
    if onlyData == None:
        return sheet_data1
    else:
        return rules
def logical_policy_def(type_ = None , file_content=None ,dev=None,file_name=None):
    returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
    data_to_return = returned[0]
    columns_ = returned[1]
    rules = save_policy(data_to_return,type_)
    sheet_name_ = 'Logical Policies'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    }    
    return sheet_data1
    
def address_def(type_ = None , file_content=None ,dev=None,file_name=None):
    returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
    data_to_return = returned[0]
    columns_ = returned[1]
    rules = save_address(data_to_return,type_,file_name)
    sheet_name_ = 'Addresses_Book'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    }
    return sheet_data1
    
def address_set_def(type_ = None , file_content=None ,dev=None,file_name=None,x=None):
    rules = []
    if x != None:
        for filee in x:
            returned =  retrive_configurations(type_ = type_ , file_content=filee ,dev=dev)
            data_to_return = returned[0]
            columns_ = returned[1]
            rule = save_address_set(data_to_return,type_,file_name)
            rules.append(rule)
    else:
        returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
        data_to_return = returned[0]
        columns_ = returned[1]
        rules = save_address_set(data_to_return,type_,file_name)
    sheet_name_ = 'Addresses_Groups'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    }     
    return sheet_data1
    
def application_def(type_ = None , file_content=None ,dev=None,file_name=None,x=None):
    rules = []
    if x != None:
        for filee in x:
            returned =  retrive_configurations(type_ = type_ , file_content=filee ,dev=dev)
            data_to_return = returned[0]
            columns_ = returned[1]
            rule = save_application(data_to_return,type_,file_name)
            rules.append(rule)
    else:
        returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
        data_to_return = returned[0]
        columns_ = returned[1]
        rules = save_application(data_to_return,type_,file_name)
    sheet_name_ = 'Applications'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    } 
    return sheet_data1
    
def application_set_def(type_ = None , file_content=None ,dev=None,file_name=None,x=None):
    rules = []
    if x != None:
        for filee in x:
            returned =  retrive_configurations(type_ = type_ , file_content=filee ,dev=dev)
            data_to_return = returned[0]
            columns_ = returned[1]
            rule = save_application_set(data_to_return,type_,file_name)
            rules.append(rule)
    else:
        returned =  retrive_configurations(type_ = type_ , file_content=file_content ,dev=dev)
        data_to_return = returned[0]
        columns_ = returned[1]
        rules = save_application_set(data_to_return,type_,file_name)
    sheet_name_ = 'APP_Groups'
    sheet_data1  = {
    'sheet_name': sheet_name_,
    'data': rules,
    'columns': columns_
    }   
    return sheet_data1


def save_excel(type_=None,word_to_find=None,cu=None, dev=None, file_content=None,file_name=None,onlyData=None):
    
    
    if type_ == 'select-node':
        sheet_data_address = []
        empty = []
        # for x in file_content:
            # sheet_data1.append(policy_def(type_ = 'policy' , file_content=x ,dev=dev,file_name=file_name))
        # logical_policy = logical_policy_def(type_ = 'logical_policy' , file_content=file_content ,dev=dev,file_name=file_name,x='y')
        for x,y in zip(file_content,file_name):
            empty = address_def(type_ = 'address' , file_content=x ,dev=dev,file_name=y)
            sheet_data_address.append(empty)
            empty = []
        # address_set = address_set_def(type_ = 'address-set' , file_content=file_content ,dev=dev,file_name=file_name,x='y')
        # application = application_def(type_ = 'application' , file_content=file_content ,dev=dev,file_name=file_name,x='y')
        # application_set = application_set_def(type_ = 'application-set' , file_content=file_content ,dev=dev,file_name=file_name,x='y')
        # sheet_data1 = [policy]
        excel_output.extract_excel(sheet_data_address,file_name)
        
    if type_ == 'all-configuration':
        policy = policy_def(type_ = 'policy' , file_content=file_content ,dev=dev,file_name=file_name)
        logical_policy = logical_policy_def(type_ = 'logical_policy' , file_content=file_content ,dev=dev,file_name=file_name)
        address = address_def(type_ = 'address' , file_content=file_content ,dev=dev,file_name=file_name)
        address_set = address_set_def(type_ = 'address-set' , file_content=file_content ,dev=dev,file_name=file_name)
        application = application_def(type_ = 'application' , file_content=file_content ,dev=dev,file_name=file_name)
        application_set = application_set_def(type_ = 'application-set' , file_content=file_content ,dev=dev,file_name=file_name)
        sheet_data1 = [address,address_set,application,application_set,policy,logical_policy]
        excel_output.extract_excel(sheet_data1,file_name)
        
    if type_ == 'policy':
        sheet_data1 = policy_def(type_ = type_ , file_content=file_content ,dev=dev,onlyData=onlyData)
        if onlyData != None:
            return sheet_data1
        else:
            sheet_na = sheet_data1['sheet_name']
            name= f'{file_name} {sheet_na}'
            excel_output.extract_excel([sheet_data1],name)
        
    if type_ == 'logical_policy':
        sheet_data1 = logical_policy_def(type_ = type_ , file_content=file_content ,dev=dev)
        sheet_na = sheet_data1['sheet_name']
        name= f'{file_name} {sheet_na}'
        excel_output.extract_excel([sheet_data1],name)
        
    if type_ == 'address':
        sheet_data1 = address_def(type_ = type_ , file_content=file_content ,dev=dev,file_name=file_name)
        sheet_na = sheet_data1['sheet_name']
        name= f'{file_name} {sheet_na}'
        excel_output.extract_excel([sheet_data1],name)
        
    if type_ == 'address-set':
        sheet_data1 = address_set_def(type_ = type_ , file_content=file_content ,dev=dev,file_name=file_name)
        sheet_na = sheet_data1['sheet_name']
        name= f'{file_name} {sheet_na}'
        excel_output.extract_excel([sheet_data1],name)
    if type_ == 'application':
        sheet_data1 = application_def(type_ = type_ , file_content=file_content ,dev=dev,file_name=file_name)
        sheet_na = sheet_data1['sheet_name']
        name= f'{file_name} {sheet_na}'
        excel_output.extract_excel([sheet_data1],name)
    
    if type_ == 'application-set':
        sheet_data1 = application_set_def(type_ = type_ , file_content=file_content ,dev=dev,file_name=file_name)
        sheet_na = sheet_data1['sheet_name']
        name= f'{file_name} {sheet_na}'
        excel_output.extract_excel([sheet_data1],name)
    
    
    


  
def matchs(match_me, each_line):
    match_results = {
        'match_system_name': re.search(r' logical-systems\s+(.*?)\s+security policies ', each_line),
        'match_system': re.search(r' logical-systems (.*?) ', each_line),
        'match_global_zone': re.search(r' policies\s+(.*?)\s+policy ', each_line),
        'match_from_zone': re.search(r' from-zone\s+(.*?)\s+to-zone ', each_line),
        'match_to_zone': re.search(r' to-zone\s+(.*?)\s+policy ', each_line),
        'match_policy_name': re.search(r' policy\s+(.*?)\s+', each_line),
        'match_address_des': re.search(r' address (.*?) description (.*?)$', each_line),
        'match_address_name': re.search(r' address (.*?) ', each_line),
        'match_address_range_ip': re.search(r' address (.*?) range-address (.*?) to (.*?)$', each_line),
        'match_address_ip': re.search(r' address (.*?) (.*?)$', each_line),
        'match_dns_address': re.search(r'dns-name (.*?)$', each_line),
        'match_address_set_name': re.search(r' address-set (.*?) address (.*?)$', each_line),
        'match_address_set_set_name': re.search(r' address-set (.*?) address-set (.*?)$', each_line),
        'match_application': re.search(r' applications application (.*?) (.*?) (.*?)$', each_line),
        'match_application_set_name': re.search(r' applications application-set (.*?) application (.*?)$', each_line),
        
        }
    return match_results.get(match_me)





def application_set_type(data_to_return,type_,file_name):
    context = []
    for each_line in data_to_return:
        match_system = matchs('match_system',each_line)
        match_application_set_name = matchs('match_application_set_name', each_line)
        if match_system:
            context.append([match_application_set_name.group(1), match_application_set_name.group(2),match_system.group(1),file_name])
        else:
            context.append([match_application_set_name.group(1), match_application_set_name.group(2),'T-Cloud', file_name])  
    return context

def application_type(data_to_return,type_,file_name):
    context = []
    for each_line in data_to_return:
        match_system = matchs('match_system',each_line)
        match_application = matchs('match_application',each_line)
        while True:
            if match_system:
                match_system_var = match_system.group(1)
            else:
                match_system_var = 'T-Cloud'
               
            found_lists = [lst for lst in context if lst[1] == match_application.group(1) and lst[5] == match_system_var]
            if not(found_lists):
                context.append(['',match_application.group(1),'','','','',match_system_var,''])
                if match_system_var == 'T-Cloud':
                    context[-1][5] ='T-Cloud'
                    if not(file_name in context[-1][7]):
                        context[-1][7] = file_name
                else:
                    context[-1][5] = match_system_var
                    if not(file_name in context[-1][7]):
                        context[-1][7] =file_name
            else:
                if match_application.group(2) == 'description':
                    context[-1][2] =match_application.group(3)
                    break
                elif match_application.group(2) == 'protocol':
                    context[-1][3] =match_application.group(3)
                    break
                elif match_application.group(2) == 'source-port':
                    context[-1][4] =match_application.group(3)
                    break
                elif match_application.group(2) == 'destination-port':
                    if re.match(r'\d-\d', match_application.group(3)):
                        context[-1][0] = 'Application Range'
                    else:
                        context[-1][0] = 'Application'
                    context[-1][5] =match_application.group(3)
                    break
    return context
    
def address_set_type(data_to_return,type_,file_name):
    context = []
    for each_line in data_to_return:
        match_system = matchs('match_system',each_line)
        match_address_set_name = matchs('match_address_set_name', each_line)
        match_address_set_set_name = matchs('match_address_set_set_name', each_line)
        if match_system:
            if match_address_set_name:
                context.append(['address-set',match_address_set_name.group(1),'address', match_address_set_name.group(2),match_system.group(1),file_name])
                continue
            if match_address_set_set_name:
                context.append(['address-set',match_address_set_set_name.group(1),'address-set', match_address_set_set_name.group(2),match_system.group(1),file_name])
                continue
        else:
            if match_address_set_name:
                context.append(['address-set',match_address_set_name.group(1),'address',match_address_set_name.group(2), 'T-Cloud',file_name])
                continue
            if match_address_set_set_name:
                context.append(['address-set',match_address_set_set_name.group(1),'address-set',match_address_set_set_name.group(2), 'T-Cloud',file_name])
                continue
                
    return context
def address_type(data_to_return,type_,file_name):
    context = []
    for each_line in data_to_return:
        match_system = matchs('match_system',each_line)
        match_address_des = matchs('match_address_des',each_line)
        match_address_name = matchs('match_address_name', each_line)
        match_address_ip = matchs('match_address_ip', each_line)
        match_address_range_ip = matchs('match_address_range_ip', each_line)
        while True:
            if match_system:
                match_system_var = match_system.group(1)
            else:
                match_system_var = 'T-Cloud'
                
            found_lists = [lst for lst in context if lst[1] == match_address_name.group(1) and lst[5] == match_system_var]
            if not(found_lists):
                context.append(['',match_address_name.group(1),'','','',match_system_var,''])
                if match_system_var == 'T-Cloud':
                    context[-1][5] ='T-Cloud'
                    
                    if not(file_name in context[-1][6]):
                        context[-1][6] = file_name
                else:
                    context[-1][5] = match_system_var
                    if not(file_name in context[-1][6]):
                        context[-1][6] =file_name
            else:
                
                if match_address_des:
                    context[-1][2] =match_address_des.group(2)
                    break
                if match_address_ip:
                    if not('range-address' in match_address_ip.group(2)):
                        if 'dns-name ' in match_address_ip.group(2):
                            match_dns_address = matchs('match_dns_address', match_address_ip.group(2))
                            context[-1][0] = 'DNS Address'
                            context[-1][3] = match_dns_address.group(1)
                            break
                        context[-1][0] = 'Address'
                        context[-1][3] =match_address_ip.group(2)
                        break
                    elif 'range-address' in match_address_ip.group(2):
                        context[-1][0] = 'Address Range'
                        context[-1][3] =match_address_range_ip.group(2)
                        context[-1][4] =match_address_range_ip.group(3)
                        break   

    return context
    
def zone_type_normal_policy(data_to_return):
    context = []
    for each_line in data_to_return:
        match_global_zone = matchs('match_global_zone',each_line)
        match_from_zone = matchs('match_from_zone', each_line)
        match_to_zone = matchs('match_to_zone', each_line)
        match_policy_name = matchs('match_policy_name', each_line)
        if match_from_zone:
            if not(['zone-based',match_from_zone.group(1),match_to_zone.group(1),match_policy_name.group(1)] in context):
                context.append(['zone-based',match_from_zone.group(1),match_to_zone.group(1),match_policy_name.group(1)])
        elif match_global_zone:
            if not([match_global_zone.group(1),'-','-',match_policy_name.group(1)] in context):
                context.append([match_global_zone.group(1),'-','-',match_policy_name.group(1)])
    return context



def zone_type_logical_policy(data_to_return):
    context = []
    for each_line in data_to_return:
        match_system_name = matchs('match_system_name',each_line)
        match_global_zone = matchs('match_global_zone',each_line)
        match_from_zone = matchs('match_from_zone', each_line)
        match_to_zone = matchs('match_to_zone', each_line)
        match_policy_name = matchs('match_policy_name', each_line)
        if match_from_zone:
            if not([match_system_name.group(1),'zone-based',match_from_zone.group(1),match_to_zone.group(1),match_policy_name.group(1)] in context):
                context.append([match_system_name.group(1),'zone-based',match_from_zone.group(1),match_to_zone.group(1),match_policy_name.group(1)])
        elif match_global_zone:
            if not([match_system_name.group(1),match_global_zone.group(1),'-','-',match_policy_name.group(1)] in context):
                context.append([match_system_name.group(1),match_global_zone.group(1),'-','-',match_policy_name.group(1)])   
    return context


def policy_type(data_to_return,type_):
    if type_ == 'policy':
        context =  zone_type_normal_policy(data_to_return)
        return [context,0]
        
    if type_ == 'logical_policy':
        context =  zone_type_logical_policy(data_to_return)
        return [context,1]  

def append_policy(data_to_return,type_,context,t):

    for each_line in data_to_return:
        match_system_name = matchs('match_system_name',each_line)
        match_global_zone = matchs('match_global_zone',each_line)
        match_from_zone = matchs('match_from_zone', each_line)
        match_to_zone = matchs('match_to_zone', each_line)
        match_policy_name = matchs('match_policy_name', each_line)
        c = 0
        for each_list in context:

            if match_from_zone:
                if each_list[t] == 'zone-based' and each_list[t+1] == match_from_zone.group(1) and each_list[t+2] == match_to_zone.group(1) and each_list[t+3] == match_policy_name.group(1):
                    
                    if ' match source-address ' in each_line:
                    
                        src_address = re.search(r' match source-address\s+(.*?)$', each_line)
                        
                        context[c][t+4] += src_address.group(1) + '\n'
                        
                    
                    if ' match destination-address ' in each_line:
                        
                        dst_address = re.search(r' match destination-address\s+(.*?)$', each_line)
                        
                        context[c][t+5] += dst_address.group(1) + '\n'
                    
                    if ' match application ' in each_line:
                        
                        app = re.search(r' match application\s+(.*?)$', each_line)
                        
                        context[c][t+6] += app.group(1) + '\n'
                        
                    if ' match dynamic-application ' in each_line:
                        
                        dyn_app = re.search(r' match dynamic-application\s+(.*?)$', each_line)
                        
                        context[c][t+7] += dyn_app.group(1) + '\n'
                        
                    if ' match url-category ' in each_line:
                        
                        url = re.search(r' match url-category\s+(.*?)$', each_line)
                        
                        context[c][t+8] += url.group(1) + '\n'
               
                    if ' then' in each_line:
                        if ' then log' in each_line:
                        
                            log_ = re.search(r' then log\s+(.*?)$', each_line)
                            
                            context[c][t+11] = log_.group(1)
                        
                        else:
                            
                            after_then = re.search(r' then\s+(.*?)$', each_line)
                            
                            
                            if 'application-services' in after_then.group(1):
                            
                                act = re.search(r' then\s+(.*?)\s+application-services', each_line)
                                
                                context[c][t+9] = act.group(1) + '\n'
                                
                                app_services = re.search(r' application-services\s+(.*?)$', each_line)
                    
                                context[c][t+10] = app_services.group(1)
                            
                            else:
                            
                                act = re.search(r' then\s+(.*?)$', each_line)
                                
                                context[c][t+9] += act.group(1) + '\n'
            
            
            elif match_global_zone:
                if each_list[t+0] == match_global_zone.group(1) and each_list[t+3] == match_policy_name.group(1):
                    if ' match source-address ' in each_line:
                    
                        src_address = re.search(r' match source-address\s+(.*?)$', each_line)
                        
                        context[c][t+4] += src_address.group(1) + '\n'
                        
                    
                    if ' match destination-address ' in each_line:
                        
                        dst_address = re.search(r' match destination-address\s+(.*?)$', each_line)
                        
                        context[c][t+5] += dst_address.group(1) + '\n'
                    
                    if ' match application ' in each_line:
                        
                        app = re.search(r' match application\s+(.*?)$', each_line)
                        
                        context[c][t+6] += app.group(1) + '\n'
                        
                    if ' match dynamic-application ' in each_line:
                        
                        dyn_app = re.search(r' match dynamic-application\s+(.*?)$', each_line)
                        
                        context[c][t+7] += dyn_app.group(1) + '\n'
                        
                    if ' match url-category ' in each_line:
                        
                        url = re.search(r' match url-category\s+(.*?)$', each_line)
                        
                        context[c][t+8] += url.group(1) + '\n'
               
                    if ' then' in each_line:
                        if ' then log' in each_line:
                        
                            log_ = re.search(r' then log\s+(.*?)$', each_line)
                            
                            context[c][t+11] = log_.group(1)
                        
                        else:
                            
                            after_then = re.search(r' then\s+(.*?)$', each_line)
                            
                            
                            if 'application-services' in after_then.group(1):
                            
                                act = re.search(r' then\s+(.*?)\s+application-services', each_line)
                                
                                context[c][t+9] = act.group(1) + '\n'
                                
                                app_services = re.search(r' application-services\s+(.*?)$', each_line)
                    
                                context[c][t+10] = app_services.group(1)
                            
                            else:
                            
                                act = re.search(r' then\s+(.*?)$', each_line)
                                
                                context[c][t+9] += act.group(1) + '\n'
        
            c+=1
    return context




def save_policy(data_to_return,type_):
    
    returned_list = policy_type(data_to_return,type_)
    context = returned_list[0]
    t = returned_list[1]
    for x in context:
            for y in range(1,9):
                x.append('')
    
    return append_policy(data_to_return,type_,context,t)
    
def save_address(data_to_return,type_,file_name):    
    rules = address_type(data_to_return,type_,file_name)
    return rules
    
    
def save_address_set(data_to_return,type_,file_name):   
    rules = address_set_type(data_to_return,type_,file_name)
    return rules
def save_application(data_to_return,type_,file_name):
    rules = application_type(data_to_return,type_,file_name)
    return rules
def save_application_set(data_to_return,type_,file_name):   
    rules = application_set_type(data_to_return,type_,file_name)
    return rules
    
def retrive_configurations(type_ = None, word_to_find = None,is_exist = 'return as list', file_content = None, dev = None):

    data_to_return = []
    
    tempelte_to_find = retrive_tempeletes(type_,word_to_find)[0]
    
    filter_ = retrive_tempeletes(type_,word_to_find)[1]
    
    columns_ = retrive_tempeletes(type_,word_to_find)[2]
    
    if dev != None: # for netconf

        data = dev.rpc.get_config(options={'format':'set'},filter_xml=filter_)
        
        retrived__data = etree.tostring(data, encoding='unicode', pretty_print=True)
        
        retrived_data = [line.strip() for line in retrived__data.strip().split('\n')]
        
        retrived_data.pop()
        
        retrived_data.pop(0)
        file_content = retrived_data
        
    if is_exist == 'yes':
    
        for each_line in file_content:
        
            if re.search(tempelte_to_find, each_line): return True
            
        return False
        
    else:

        for each_line in file_content:
        
            if re.search(tempelte_to_find, each_line):
                
                data_to_return.append(each_line)
            
        return [data_to_return,columns_]

    
def retrive_tempeletes(type_,word_to_find_=None):

    if type_ == 'policy':
    
        match = r'set security policies '
        
        filter_ = 'security/policies'
        
        columns_ = ['Zone Type','From Zone','To Zone', 'Policy Name', 'Source Address', 'Destination Address' ,'Application','Dynamic Application','URL Category','Action','Application Services','Log']

        return [match,  filter_,columns_]
        
    if type_ == 'logical_policy':
    
        match = r'set logical-systems .* security policies '
        
        filter_ = 'security/policies'
        
        columns_ = ['System Name','Zone Type','From Zone','To Zone', 'Policy Name', 'Source Address', 'Destination Address' ,'Application','Dynamic Application','URL Category','Action','Application Services','Log']

        return [match,  filter_,columns_]    
    if type_ == 'address':
    
        match = r' security address-book global address '
        
        filter_ = 'security/address-book'
        
        columns_ = ['Type','Address Name', 'Address Description', 'Address Value', 'End Address Range' ,'Location', 'Model']

        return [match,  filter_,columns_] 
        
    if type_ == 'address-set':
    
        match = r' security address-book global address-set '
        
        filter_ = 'security/address-book'
        
        columns_ = ['Address Type','Group Name','Group Members','Address Type','Location', 'Model']

        return [match,  filter_,columns_] 
        
    if type_ == 'application':
    
        match = r' applications application '
        
        filter_ = 'applications'
        
        columns_ = ['Type','Service_Name','Service_Description','Service_Protocol', 'Service_Src-Port', 'Service_DST-Port', 'Location', 'Model']

        return [match,  filter_,columns_] 
        
    if type_ == 'application-set':
    
        match = r' applications application-set '
        
        filter_ = 'applications'
        
        columns_ = ['APP_Group_Name','APP_Group_Members','Location', 'Model']

        return [match,  filter_,columns_]         
    
    
 








import load_full_set_configurations
import _main_functions_
import re
from srx_classes import all_applications_type
from srx_classes import all_policy_type
from srx_classes import all_address_type
from lxml import etree
import clipboard
from termcolor import colored


def load_policy(cu=None,dev=None,file_content=None):
    print('►►►►►►►►►►►►►►►►► MAKE SURE THE Policies IN THE CLIPBOARD ◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄')
    input('Press any key to continue...')
    print('▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼ Paste your set commands ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼')
    lines = clipboard.paste()

    
    get_set_commands = lines.strip().split('\n')   

    result = [line.replace('\r','').split('\t') for line in get_set_commands]
    for x in result:print(x)
    print(f'▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲           Count = {len(result)}           ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')
    user_agrees = input('type y to continue : ')
    if user_agrees.lower() == 'y':
        
        #list_of_list = [commands, invalid_application_set, application_description,valid_application_count]

        list_of_list = all_policy_type.policy_set( result,dev=dev,file_content=file_content)
    else:
        return True
    
            
    if len(list_of_list[0]) > 0:
        print(g('▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼  ') + y('Valid Commands.') + g('  ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼'))
        for c in list_of_list[0]:
            print(c)
        print(g('▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲  ') + y('End of Valid Commands.') + g('  ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲'))
    
    
    if dev != None :
        pushing(cu,list_of_list[0],list_of_list[1])

    return True

def load_application(cu=None,dev=None,file_content=None):
    print('►►►►►►►►►►►►►►►►► MAKE SURE THE COMMANDS IN THE CLIPBOARD ◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄')
    input('Press any key to continue...')
    print('▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼ Paste your set commands ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼')
    lines = clipboard.paste()
    get_set_commands = lines.strip().split('\n')
    get_set_commands = [element.replace('\r', '').strip() for element in get_set_commands]
    for x in get_set_commands:print(x)
    print(f'▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲           Count = {len(get_set_commands)}           ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')
    user_agrees = input('type y to continue : ')
    if user_agrees.lower() == 'y':
        
        #list_of_list = [commands, invalid_application_set, application_description,valid_application_count]
        list_of_list = all_applications_type.application_applications_set(dev, lines,file_content=file_content)
    else:
        return True
    if list_of_list[3] > 0:
        print(f'********************* Valid Application. Count : {list_of_list[3]} *********************')
        for c in list_of_list[0]:
            print(c)
            
    if len(list_of_list[1]) > 0: 
        print(f'********************* Invalid Application. Count : {len(list_of_list[1])} *********************')
        for c in list_of_list[1]:
            print(c)
            
    if len(list_of_list[4]) > 0:     
        print(f'********************* Duplicated Input. Count : {len(list_of_list[4])} *********************')
        for c in list_of_list[4]:
            print(c)
            
    if len(list_of_list[5]) > 0:     
        print(f'********************* Already Excisted. Count : {len(list_of_list[5])} *********************')
        for c in list_of_list[5]:
            print(c)
    
    
    if dev != None :
        pushing(cu,list_of_list[0],list_of_list[2])

    return True



def load_adresses(cu=None,dev=None,file_content=None):
    print('►►►►►►►►►►►►►►►►► MAKE SURE THE COMMANDS IN THE CLIPBOARD ◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄')
    input('Press any key to continue...')
    print('▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼ Paste your set commands ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼')
    lines = clipboard.paste()
    get_set_commands = lines.strip().split('\n')
    get_set_commands = [element.replace('\r', '').strip() for element in get_set_commands]
    for x in get_set_commands:print(x)
    print(f'▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲           Count = {len(get_set_commands)}           ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')
    user_agrees = input('type y to continue : ')
    if user_agrees.lower() == 'y':
        #list_of_list = [commands, wrong_inputs, invalid_range, invalid_address_set, address_description]
        list_of_list = all_address_type.address_address_set(dev, lines,file_content)
    else:
        return True
    if len(list_of_list[0]) > 0:         
        print(f'********************* Valid Commands Count : {len(list_of_list[0])} *********************')
        for c in list_of_list[0]:
            print(c)
      
    if len(list_of_list[1]) > 0:         
        print(f'********************* Wrong Inputs Count : {len(list_of_list[1])} *********************')
        for c in list_of_list[1]:
            print(c)
        
    if len(list_of_list[2]) > 0:         
        print(f'********************* Invalid Range Count : {len(list_of_list[2])} *********************')
        for c in list_of_list[2]:
            print(c)
        
    if len(list_of_list[3]) > 0:         
        print(f'********************* Invalid Address Set Count : {len(list_of_list[3])} *********************')
        for c in list_of_list[3]:
            print(c)
        
    if len(list_of_list[5]) > 0:     
        print(f'********************* Duplicated Input. Count : {len(list_of_list[5])} *********************')
        for c in list_of_list[5]:
            print(c)
            
    if len(list_of_list[6]) > 0:     
        print(f'********************* Already Excisted. Count : {len(list_of_list[6])} *********************')
        for c in list_of_list[6]:
            print(c)    
            
    if len(list_of_list[7]) > 0:     
        print(f'********************* The IP Already Excisted In Configuration. Count : {int(len(list_of_list[7])/2)} *********************')
        for c in list_of_list[7]:
            print(c)
    
    if len(list_of_list[8]) > 0:     
        print(f'********************* The Same IP You Paste Twice. Count : {len(list_of_list[8])} *********************')
        for c in list_of_list[8]:
            print(c)
            
    if dev != None :
        pushing(cu,list_of_list[0],list_of_list[4])
    
    
    return True
        
        
    
    
        
    
    
                
                
                
def load_full_set_commands(cu):
        print('►►►►►►►►►►►►►►►►► MAKE SURE THE COMMANDS IN THE CLIPBOARD ◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄◄')
        input('Press any key to continue...')
        print('▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼ Paste your set commands ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼')
        lines = clipboard.paste()
        get_set_command = lines.strip().split('\n')
        get_set_commands = [element.replace('\r', '').strip() for element in get_set_command]
        for x in get_set_commands:print(x)
        print(f'▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲           Count = {len(get_set_commands)}           ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')
        user_agrees = input('type y to continue : ')
        if user_agrees.lower() == 'y':
            load_full_set_configurations.load_list_set_configurations(cu,get_set_commands) #loading the conf
        else:
            return True
        
        _main_functions_.commit_check(cu)
        
        _main_functions_.commit(cu)
        
        return True


def pushing(cu, configuration_list, configuration_decription):
    user_agrees = input('type y to continue : ')
    if user_agrees.lower() != 'y':return True  
    
    load_full_set_configurations.load_list_set_configurations(cu,configuration_list)
    
    _main_functions_.commit_check(cu)
    
    _main_functions_.commit(cu,configuration_decription)
    
    
    
    
def retrive_configurations(type_, word_to_find, is_exist = 'return as list', file_content = None, dev = None):

    data_to_return = []
    
    tempelte_to_find = re.sub(r'\x1b\[[0-9;]+m\x1b\[[0-9;]+m|\x1b\[[0-9;]+m', '', retrive_tempeletes(type_,word_to_find)[0]) 
    
    filter_ = retrive_tempeletes(type_,word_to_find)[1]
    
    if dev != None: # for netconf

        data = dev.rpc.get_config(options={'format':'set'},filter_xml=filter_)
        
        retrived__data = etree.tostring(data, encoding='unicode', pretty_print=True)
        
        retrived_data = [line.strip() for line in retrived__data.strip().split('\n')]
        
        retrived_data.pop()
        
        retrived_data.pop(0)
        
        file_content = retrived_data

    if is_exist == 'yes':
    
        for each_line in file_content:
            
            if re.search(tempelte_to_find, each_line): return True
            
        return False
        
    else:
    
        for x in file_content:
            each_line = re.sub(r'\x1b\[[0-9;]+m\x1b\[[0-9;]+m|\x1b\[[0-9;]+m', '', x)
            if re.search(tempelte_to_find, each_line):

                data_to_return.append(each_line)
            
        return data_to_return


def retrive_tempeletes(type_,word_to_find_):

    if type_ == 'address':
        pattern = r'security address-book global address {} '.format(word_to_find_)
        return [pattern,  'security/address-book']

    if type_ == 'address-set':
        pattern = r'security address-book global address-set {} address {}'.format(word_to_find_[0],word_to_find_[1])
        return [pattern,  'security/address-book']
    
    if type_ == 'address-ip': 
        pattern = r'security address-book global address .* {}'.format(word_to_find_)
        return [pattern,  'security/address-book']
        
    if type_ == 'address-range-ip':
        pattern = r'security address-book global address .* range-address {} to {}'.format(word_to_find_[0],word_to_find_[1])
        return [pattern,  'security/address-book']
    
    if type_ == 'Tcloud_policy_address-set':
        pattern = r'set security address-book global address-set {} address (.*?)$'.format(word_to_find_)
        return [pattern,  'security/address-book']
    
    if type_ == 'Tcloud_policy_address': 
        pattern = r'set security address-book global address {} (.*?)$'.format(word_to_find_)
        return [pattern,  'security/address-book']
    
    if type_ == 'Tcloud_policy_app-set':
        pattern = r'set applications application-set {} application (.*?)$'.format(word_to_find_)
        return [pattern,  'applications']
    
    if type_ == 'Tcloud_policy_app': 
        pattern = r'set applications application {} protocol (.*?)$'.format(word_to_find_)
        return [pattern,  'applications']
    
    if type_ == 'Tcloud_idp': 
        pattern = r'set system security-profile LSYS-Resoreses-Profile idp-policy IDP_Default$'
        return [pattern,  'system/idp']
    
    if type_ == 'all': 
        pattern = r'set'
        return [pattern,  'security']
            
    if type_ == 'application':
        pattern = r'application {} protocol '.format(word_to_find_)
        return [pattern, 'applications']
    
    if type_ == 'application-set':
        pattern = r'application-set {} application {}'.format(word_to_find_[0],word_to_find_[1])
        return [f'application-set {word_to_find_[0]} application {word_to_find_[1]}', 'applications']
    

  
    
def m(word):
    styled_text = colored(word, "magenta", attrs=["bold"])
    return styled_text
def c(word):
    styled_text = colored(word, "cyan", attrs=["bold"])
    return styled_text
def b(word):
    styled_text = colored(word, "blue", attrs=["bold"])
    return styled_text
def y(word):
    styled_text = colored(word, "yellow", attrs=["bold"])
    return styled_text
def g(word):
    styled_text = colored(word, "green", attrs=["bold"])
    return styled_text


import re
import ipaddress
from termcolor import colored
from colorama import init, Fore
from srx_classes import _configurations_
from srx_classes import all_address_type
from srx_classes import save_excel_def

# Initialize colorama
init()
policy_description = 'Policy Pushed through Netconf'

include_app = [
    'junos-icmp-all',
    ]


def make_new_address(address,ip,dev=None,file_content=None):
    to = ' to '
    if ip.find(to) != -1:
        send_address=f'address\t{address}\t\t{ip.split(to)[0]}\t{ip.split(to)[1]}' 
    else:
        send_address=f'address\t{address}\t\t{ip}' 
        
    list_of_list = all_address_type.address_address_set(dev,send_address,file_content=file_content)
    

    if len(list_of_list[0]) > 0:         
        return list_of_list[0]
      
    if len(list_of_list[1]) > 0:         
        print(yellow('!!!!!!!! Wrong Input, Please try again !!!!!!!!'))
        return 'try again'
        
    if len(list_of_list[2]) > 0:  
        print(yellow('!!!!!!!! Invalid Range, Please try again !!!!!!!!'))       
        return 'try again'
        
    if len(list_of_list[3]) > 0: 
        print(yellow('!!!!!!!! Invalid Address, Please try again !!!!!!!!'))        
        return 'try again'
        
    if len(list_of_list[5]) > 0:     
        print(yellow('!!!!!!!! Duplicated Input, Please try again !!!!!!!!'))
        return 'try again'
            
    if len(list_of_list[6]) > 0:     
        print(yellow('!!!!!!!! Already Excisted, Please try again !!!!!!!!'))
        return 'try again'   
            
    if len(list_of_list[7]) > 0:    
        print(yellow('!!!!!!!! The IP Already Excisted In Configuration, Please try again !!!!!!!!')) 
        return 'try again'
    
    if len(list_of_list[8]) > 0:     
        print(yellow('!!!!!!!! The Same IP You Paste Twice, Please try again !!!!!!!!'))
        return 'try again'
        
        
def find_address_caseSensitve(address,dev=None,file_content=None):
    if dev != None:
        all_set = _configurations_.retrive_configurations('all' ,address, dev=dev)
    if file_content != None:
        all_set = _configurations_.retrive_configurations('all' ,address, file_content=file_content)
    
    for line in all_set:
        match = re.match(r'set security address-book global address (.*?) (.*?)$',line)
        if match:
            configure_address_name = re.match(r'set security address-book global address (.*?) (.*?)$',line).group(1)
            configure_address_ip = re.match(r'set security address-book global address (.*?) (.*?)$',line).group(2)
            if is_valid_ipv4(configure_address_ip):
                if address.lower() == configure_address_name.lower():
                    return [configure_address_name,configure_address_ip]
                else:
                    return 'address not found'
    


def get_address_ip(address,dev=None,file_content=None):
    if dev != None:
        retuen_match_in_list_set = _configurations_.retrive_configurations('Tcloud_policy_address-set' , address, dev=dev)
        retuen_match_in_list = _configurations_.retrive_configurations('Tcloud_policy_address' , address, dev=dev)
    if file_content != None:
        retuen_match_in_list_set = _configurations_.retrive_configurations('Tcloud_policy_address-set' , address, file_content=file_content)
        retuen_match_in_list = _configurations_.retrive_configurations('Tcloud_policy_address' , address, file_content=file_content)
    address = re.sub(r'\x1b\[[0-9;]+m\x1b\[[0-9;]+m|\x1b\[[0-9;]+m', '', address)
    if len(retuen_match_in_list_set)== 0 : # not found as address set 
        if len(retuen_match_in_list) == 0 :# not found as address
            return 'address not found'
        else: # found as address
            match1 = re.search(r'address (.*?) (.*?)$',retuen_match_in_list[0])
            if 'range-address' in match1.group(2):
                match1 = re.search(r'address {} range-address (.*?) to (.*?)$'.format(address),retuen_match_in_list[0])
                return match1.group(1)
            else:
                ip = ip_without_prefix(match1.group(2))
                return ip
            
    else:#found as address set 
        add = re.search('address (.*?)$',retuen_match_in_list_set[-1]).group(1)
        if dev != None:
            retuen_match_in_list = _configurations_.retrive_configurations('Tcloud_policy_address' , add, dev=dev)
        if file_content != None:
            retuen_match_in_list = _configurations_.retrive_configurations('Tcloud_policy_address' , add, file_content=file_content)
        if len(retuen_match_in_list) == 0 :# not found as address
            return 'address not found'
        else: # found as address
            match1 = re.search(r'address (.*) (.*?)$',retuen_match_in_list[-1])
            if match1:
                if 'range-address' in match1.group(2):
                    match1 = re.search(r'address {} range-address (.*?) to (.*?)$'.format(address),retuen_match_in_list[0])
                    return match1.group(1)
                else:
                    ip = ip_without_prefix(match1.group(2))
                    return ip
            else:
                return 'address not found'
        
        
        
def get_zone_name(ip,dev=None,file_content=None): 
    if dev != None:
        retuen_match_in_list_route = _configurations_.retrive_configurations('all' ,ip, dev=dev)
    if file_content != None:
        retuen_match_in_list_route = _configurations_.retrive_configurations('all' ,ip, file_content=file_content)
    
    specific_range = []
    
    the_matched_commands = []

    r_specific_range = []
    
    r_the_matched_commands = []
    while True:
        for each_ip in retuen_match_in_list_route:
            match_unit_address = re.match(r'set interfaces (.*?) unit (.*?) family inet address (.*?)$',each_ip)
            if match_unit_address:
                interface_IP = re.match(r'set interfaces (.*?) unit (.*?) family inet address (.*?)$',each_ip).group(3)
                if is_ip_in_range(ip,interface_IP):
                    specific_range.append(interface_IP)
                    the_matched_commands.append(each_ip)
            else:
                match_route_address_ = re.match(r'set routing-options static route (.*?) next-hop (.*?)$',each_ip)
                if match_route_address_:
                    route_IP_ = re.match(r'set routing-options static route (.*?) next-hop (.*?)$',each_ip).group(1)
                    if is_ip_in_range(ip,route_IP_):
                        r_specific_range.append(route_IP_)
                        r_the_matched_commands.append(each_ip)
                        
        else:
            if len(specific_range) >= len(r_specific_range):
                break
            elif len(r_specific_range) > len(specific_range):
                ip_index = most_specific_range(ip,r_specific_range)
                for x in r_the_matched_commands:
                    ipz = re.match(r'set routing-options static route {} next-hop (.*?)$'.format(ip_index),x)
                    if ipz:
                        ip = re.match(r'set routing-options static route {} next-hop (.*?)$'.format(ip_index),x).group(1)
                        specific_range = []
                        r_specific_range = []



    command_index = most_specific_range(ip,specific_range)
    
    for x in the_matched_commands:
        ipc = re.match(r'set interfaces (.*?) unit (.*?) family inet address {}$'.format(command_index),x)
        if ipc :
            interface = re.match(r'set interfaces (.*?) unit (.*?) family inet address {}$'.format(command_index),x).group(1)
            unit = re.match(r'set interfaces (.*?) unit (.*?) family inet address {}$'.format(command_index),x).group(2)
            vlan_name = f'{interface}.{unit}'   
    
    for each_interface in retuen_match_in_list_route:
        match_interface = re.match(r'set security zones security-zone (.*?) interfaces {}$'.format(vlan_name),each_interface)
        if match_interface:
            interface_zone = re.match(r'set security zones security-zone (.*?) interfaces {}$'.format(vlan_name),each_interface).group(1)
            return interface_zone
                

def get_configured_policy(file_content=None):
    data = save_excel_def.save_excel(type_='policy',file_content=file_content, file_name='no name',onlyData='data')
    return data

def print_same_context_same_ruleName(old_rule_name,src_zone,dst_zone,three_values,new_rule_name,src_address,dst_address,app_Group):
    n = '\n'
    a = ' & '
    j = ', '
    e = ''
    z = 'zone-type'
    New_Src = 'New Src. Add'
    New_Dst = 'New Dst. Add'
    New_Service= 'New Service'

    print(yellow('!!!!!!!!! WARNING Duplicated Rule!!!!!!!!!'))
    print(g('▼▼▼▼▼▼▼▼▼▼▼▼ [')+b(new_rule_name)+g('] similar to an Existing rule called [')+b(old_rule_name)+g('].'))
    print(f' ►► Zone Type   : {b(z)}')
    print(f' ►► From Zone   : {b(src_zone)}')
    print(f' ►► To Zone     : {b(dst_zone)}')
    print(f' ►► Rule Name   : {b(old_rule_name)} ►► {b(new_rule_name)}')
    print(f' ►► Existing Src. Add    : {j.join([b(x) if x in src_address else x for x in three_values[0] if x != e])} ')
    print(f' ►► {g(New_Src)}         : {j.join([b(x) if x in three_values[0] else x for x in src_address if x != e])} ')
    print(f' ►► Existing Dst. Add    : {j.join([b(x) if x in dst_address else x for x in three_values[1] if x != e])} ')
    print(f' ►► {g(New_Dst)}         : {j.join([b(x) if x in three_values[1] else x for x in dst_address if x != e])} ')
    print(f' ►► Existing Service     : {j.join([b(x) if x in app_Group else x for x in three_values[2] if x != e])} ')
    print(f' ►► {g(New_Service)}          : {j.join([b(x) if x in three_values[2] else x for x in app_Group if x != e])} ')
    print('▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲ End Of the rule ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')

def print_rule(old_rule,rule_name,Sources,Destinations,Services):
    n = '\n'
    a = ' & '
    j = ', '
    e = ''
    
    New_Src = 'New Src. Add'
    New_Dst = 'New Dst. Add'
    New_Service= 'New Service'

    print(yellow('!!!!!!!!! WARNING Duplicated Rule Name !!!!!!!!!'))
    print(g('▼▼▼▼▼▼▼▼▼▼▼▼ [')+m(rule_name)+g('] similar to an Existing rule called [')+m(old_rule[3])+g('].'))
    print(f' ►► Zone Type   : {old_rule[0].replace(n,a)}')
    print(f' ►► From Zone   : {old_rule[1]}')
    print(f' ►► To Zone     : {old_rule[2]}')
    print(f' ►► Rule Name   : {b(old_rule[3])} ►► {b(rule_name)}')
    print('--------------------------:------------------------------------------------------')
    print(f' ►► Existing Src. Add    : {j.join([b(x) if x in Sources else x for x in old_rule[4].split(n) if x != e])} ')
    print(f' ►► {g(New_Src)}         : {j.join([b(x) if x in old_rule[4].split(n) else x for x in Sources if x != e])} ')
    print('--------------------------------------------------------------------------------')
    print(f' ►► Existing Dst. Add    : {j.join([b(x) if x in Destinations else x for x in old_rule[5].split(n) if x != e])} ')
    print(f' ►► {g(New_Dst)}         : {j.join([b(x) if x in old_rule[5].split(n) else x for x in Destinations if x != e])} ')
    print('--------------------------:------------------------------------------------------')
    print(f' ►► Existing Service     : {j.join([b(x) if x in Services else x for x in old_rule[6].split(n) if x != e])} ')
    print(f' ►► {g(New_Service)}          : {j.join([b(x) if x in old_rule[6].split(n) else x for x in Services if x != e])} ')
    print('--------------------------:------------------------------------------------------')
    print(f' ►► Dyn. App    : {old_rule[7]}')
    print(f' ►► Url. Cat    : {old_rule[8]}')
    print(f' ►► Action      : {j.join([x for x in old_rule[9].split(n) if x != e])}')
    print(f' ►► App. Serv   : {old_rule[10]}')
    print(f' ►► Log         : {old_rule[11]}')
    print('▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲ End Of the rule ▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲▲')



def  get_app(each_app, dev = None, file_content = None):
    if dev != None:
        retuen_match_set = _configurations_.retrive_configurations('Tcloud_policy_app-set' , each_app,is_exist = 'yes', dev=dev)
        retuen_match = _configurations_.retrive_configurations('Tcloud_policy_app' , each_app,is_exist = 'yes', dev=dev)
    if file_content != None:
        retuen_match_set = _configurations_.retrive_configurations('Tcloud_policy_app-set' , each_app,is_exist = 'yes', file_content=file_content)
        retuen_match = _configurations_.retrive_configurations('Tcloud_policy_app' , each_app,is_exist = 'yes', file_content=file_content)

    if retuen_match == True or retuen_match_set == True:
        return True
    else:
        return False

def check_for_valid_app_format(each_app): # [True,protocol,port]

    parts = each_app.split('_')
    port = parts[1].split('-')
    if len(port) == 1 and port[0].isdigit():
        return [True,parts[0],parts[1]]
        
    elif len(port) == 2 and port[0].isdigit() and port[1].isdigit():
        return [True,parts[0],parts[1]]
    else:
        return [False,False,False]

    
def create_new_app(protocol,port): #['app',['command1','command2']] or [False,False]
    app = f'{protocol.upper()}_{port}'
    command=[
        f'set applications application {m(app)} protocol {m(protocol.lower())}',
        f'set applications application {m(app)} destination-port {m(port)}',
    ]
    return [app,command]

def return_application(Services,dev=None,file_content=None):
    ser_com_LIST = []
    for each_app in Services:
        app = get_app(each_app, dev = dev, file_content = file_content)
        if app or each_app in include_app: #if app found
            ser_com_LIST.append([each_app,''])
            
        else: # if app not found
            # each_app = TCP_443 or TCP_123-233 or ser_group
            print(g('This Application [')+m(each_app)+g('] Not Found.Do you want to create it.'))
            user_agrees = input('type ['+yellow('y')+'] to continue : ')
            if user_agrees.lower() == 'y':
                after_check = check_for_valid_app_format(each_app) # [True,protocol,port]
                if after_check[0]:
                    protocol = after_check[1]
                    port = after_check[2]
                    created_app = create_new_app(protocol,port) #['app',['command1','command2']] or [False,False]
                    ser_com_LIST.append(created_app)
                else:
                   print(g('You can not make application-set in this section. This Application will be skipped'))
                   ser_com_LIST.append([False,False]) 
            else:
                print(g('This Application will be skipped'))
                ser_com_LIST.append([False,False])
    return  ser_com_LIST             

def return_zones(addresses,dev=None,file_content=None): 
    # return #[[address,zone,[command]],[False,False,False]]   # command may == ''  
    add_zon_com_LIST = []
    
    for count,each_source in enumerate(addresses):
        flag = True
        while True:
            ip = get_address_ip(each_source,dev=dev,file_content=file_content) # the ip of the given address
            if ip != 'address not found': # if ip found
                zone_name = get_zone_name(ip,dev=dev,file_content=file_content)
                add_zon_com_LIST.append([each_source,zone_name,''])
                break
            else:# if ip not found
                
                # try find somthing similar
                found_address = find_address_caseSensitve(each_source,dev=dev,file_content=file_content)
                
                #not fount but found similr
                if found_address != 'address not found' and found_address !=None:
                    print(g('This address [')+m(each_source)+g('] Not Found.Did you mean [')+m(found_address[0])+g('] with IP [')+m(found_address[1])+g('].'))
                    user_agrees = input('type ['+yellow('y')+'] to continue : ')
                    if user_agrees.lower() == 'y':
                        addresses[count] = found_address[0]
                        each_source = found_address[0]
                        continue
                    else:
                        # if you not want to select the old try make new one
                        print(g('Do you want to create this Address  [') +  m(each_source) + g('].'))
                        user_agrees = input('type ['+yellow('y')+'] to continue : ')
                        if user_agrees.lower() == 'y':
                            while flag:
                                user_input = input(g('Enter [') + m(each_source) + g('] Ip or Range : '))
                                commands = make_new_address(each_source,user_input,dev=dev,file_content=file_content)
                                if commands != 'try again':
                                    ip = get_address_ip(each_source,dev=dev,file_content=commands) # the ip of the given address
                                    if ip != 'address not found': # if ip found
                                        zone_name = get_zone_name(ip,dev=dev,file_content=file_content)
                                        add_zon_com_LIST.append([each_source,zone_name,commands])
                                        flag = False
                            break
                                        
                            
                        else:
                            add_zon_com_LIST.append([False,False,False])
                            break
                            
                else:# not found at all
                    print(g('This address [')+m(each_source)+g('] Not Found.Do you want to create it.'))
                    user_agrees = input('type ['+yellow('y')+'] to continue : ')
                    if user_agrees.lower() == 'y':
                        while flag:
                            user_input = input(g('Enter [') + m(each_source) + g('] Ip or Range : '))
                            commands = make_new_address(each_source,user_input,dev=dev,file_content=file_content)
                            if commands != 'try again':
                                color_address = f'\x1b[1m\x1b[35m{each_source}\x1b[0m' 
                                ip = get_address_ip(color_address,dev=dev,file_content=commands) # the ip of the given address
                                if ip != 'address not found': # if ip found
                                    zone_name = get_zone_name(ip,dev=dev,file_content=file_content)
                                    add_zon_com_LIST.append([each_source,zone_name,commands])
                                    flag = False
                        break
                                    
                        
                    else:
                        add_zon_com_LIST.append([False,False,False])
                        break
                            
                            
    return  add_zon_com_LIST                       

def commands_zonesGroup(Address_Zones,Address_Zones_Group):
    commands=[]
    for x in Address_Zones:
        if x != [False,False,False]:
            if x[2] != '' and x[2] != False:
                commands.append(x[2])
            for index,each_group in enumerate(Address_Zones_Group):
                if x[1] in each_group :
                    if not(x[0] in each_group[1]):
                        Address_Zones_Group[index][1].append(x[0])
                        break
            else:
                Address_Zones_Group.append([x[1],[x[0]]])
    commands.append('')
    return [commands,Address_Zones_Group]

def commands_appGroup(applications):#[['app',['command1','command2']],[False,False]]  # command may == '' 
    app=[]
    commands=[]
    for x in applications:
        if x != [False,False]:
            if x[1] != '' and x[1] != False:
                for cmd in x[1]:
                    commands.append(cmd)

            if x[0] != '' and x[0] != False:
                app.append(x[0])
    commands.append('')
    return [commands,app]

def True_if_idp(dev=None, file_content = None):
    if dev != None:
        return _configurations_.retrive_configurations('Tcloud_idp' , 'line',is_exist = 'yes', dev=dev)
    if file_content != None:
        return _configurations_.retrive_configurations('Tcloud_idp' , 'line',is_exist = 'yes', file_content=file_content)

def same_context_same_ruleName(rule_name=None,context=None,src_address=None,dst_address=None,dev=None, file_content = None):
    if dev != None:
        all_set = _configurations_.retrive_configurations('all' ,'address', dev=dev)
    if file_content != None:
        all_set = _configurations_.retrive_configurations('all' ,'address', file_content=file_content)
    matched_lines = []
    for line in all_set:
        if f'set security policies {context} policy {rule_name} ' in line:
            matched_lines.append(line)
    return matched_lines

def same_context_same_adresses(rule_name=None,context=None,src_address=None,dst_address=None,dev=None, file_content = None):
    if dev != None:
        all_set = _configurations_.retrive_configurations('all' ,'address', dev=dev)
    if file_content != None:
        all_set = _configurations_.retrive_configurations('all' ,'address', file_content=file_content)
    
    matched_lines = []
    policy_name= []
    match_policy_context = []
    matched_policy_name = []

    for line in all_set:
        match_me = re.match(r'set security policies {} policy (.*?) '.format(context),line)
        if match_me:
            ruleName = match_me.group(1)
            policy_name.append(ruleName)
            for index,x in enumerate(match_policy_context):
                if x[0]== policy_name:
                    if not(line in match_policy_context[index][1]):
                        match_policy_context[index][1].append(line)
            else:
                match_policy_context.append([policy_name,[line]])

    for each_policy in match_policy_context:
        for add in src_address:
            if not(f'set security policies {context} policy {each_policy[0]} source-address {add}' in each_policy[1]):
                break
        else:
            if not(each_policy[0] in matched_policy_name):
                matched_policy_name.append(each_policy[0])

    for each_policy in match_policy_context:
        for add in dst_address:
            if not(f'set security policies {context} policy {each_policy[0]} destination-address {add}' in each_policy[1]):
                break
        else:
            if not(each_policy[0] in matched_policy_name):
                matched_policy_name.append(each_policy[0])        

    for x in match_policy_context:
            if x[0] in matched_policy_name:
                matched_lines.append(x)

    return matched_lines

def get_values_from_commands(lines):
    src=[]
    dst=[]
    ser=[]
    for each_line in lines:
        if re.match(r'(.*?) source-address (.*?)$',each_line):
            src.append(re.match(r'(.*?) source-address (.*?)$',each_line).group(2))
        if re.match(r'(.*?) destination-address (.*?)$',each_line):
            dst.append(re.match(r'(.*?) destination-address (.*?)$',each_line).group(2))
        if re.match(r'(.*?) application (.*?)$',each_line):
            ser.append(re.match(r'(.*?) application (.*?)$',each_line).group(2))
    return [src,dst,ser]

def print_same_zone_error(rule_name,each_src,each_dst,src_address,dst_address,in_,not_):
    print('----------------------------This rule had been skipped----------------------------')
    print(f'This Rule : {m(rule_name)} must be in {in_} Not in {not_} because both zones in {in_}')
    print(f'Source Zone : {m(each_src)}')
    print(f'Destination Zone : {m(each_dst)}')
    print(f'Source Adress : {yellow(src_address)}')
    print(f'Destination Adress : {yellow(dst_address)}')
    print('----------------------------This rule had been skipped----------------------------')

def policy_maker(rule_name,Src_Zon_Group,Dst_Zon_Group,app_Group,dev=None ,file_content=None):
#[[zone1,[add1,add2]],[zone2,[add3,add4]]]
#[[zone3,[add5,add6]],[zone4,[add7,add8]],[zone5,[add9,add10]]]
#['app1','app2','app3'],
    commands=[]
    for each_src in Src_Zon_Group:
        for each_dst in Dst_Zon_Group:
            context = f'from-zone {each_src[0]} to-zone {each_dst[0]}'
            src_address = [src for src in each_src[1]]
            dst_address = [dst for dst in each_dst[1]]
            
            #['set sec..','set sec..','set sec..']
            check_Context_RuleName = same_context_same_ruleName(rule_name=rule_name,context=context,dev=dev, file_content = file_content)

            if len(check_Context_RuleName) > 0 :
                three_values = get_values_from_commands(check_Context_RuleName)
                print_same_context_same_ruleName(rule_name,each_src[0],each_dst[0],three_values,rule_name,src_address,dst_address,app_Group)
            
            #[[policyName,['set sec..','set sec..','set sec..']],[policyName,['set sec..','set sec..','set sec..']]]
            check_Context_address = same_context_same_adresses(context=context,src_address=src_address,dst_address=dst_address,dev=dev, file_content = file_content)

            if len(check_Context_address) > 0 :
                for x in check_Context_address:
                    thre_values = get_values_from_commands(x[1]) 
                    print_same_context_same_ruleName(x[0],each_src[0],each_dst[0],thre_values,rule_name,src_address,dst_address,app_Group)
            
            if each_src[0] == each_dst[0] =='Outside_to_DCFW':
                print_same_zone_error(rule_name,each_src[0],each_dst[0],src_address,dst_address,'DFW','PFW')
                break

            if each_src[0] == each_dst[0] =='Management_Platform_FW':
                print_same_zone_error(rule_name,each_src[0],each_dst[0],src_address,dst_address,'PFW','DFW')
                break
            
            if each_src[0] == each_dst[0] =='DC_FW_to_ACI_North-South':
                 each_src[0] = each_dst[0] = 'DC_FW_to_ACI_East-West'
            
            for src_add in each_src[1]:
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} match source-address {m(src_add)}')

            for dst_add in each_dst[1]:
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} match destination-address {m(dst_add)}')

            for app in app_Group:
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} match application {m(app)}')

            if True_if_idp(dev=dev ,file_content=file_content):
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} match dynamic-application ' + m('any'))
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} match url-category '+ c('none'))
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} then permit application-services idp-policy Recommended')
            else:
                commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} then ' + c('permit'))
            commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} then log ' + c('session-close'))
            commands.append(f'set security policies from-zone {m(each_src[0])} to-zone {m(each_dst[0])} policy {m(rule_name)} then ' + c('count'))
    commands.append('')
    return commands

def policy_set(pasted_from_cli, dev=None, file_content = None):
    
    skipped_rules = []
    final_commands=[]
    invalid_address = []
    for each_line in pasted_from_cli:
    
        rule_name = each_line[3]
        Sources = values_after(each_line,'Src') #Sources = [Sources1,Sources2,Sources3]
        Destinations = values_after(each_line,'Dst')#Destinations = [Destinations1,Destinations2,Destinations3]
        Services = values_after(each_line,'Ser')#Services = [Services1,Services2,Services3]
        Source_Zones = []
        Source_Zones_Group = []
        Destination_Zones = []
        Destination_Zones_Group = []
        applications = []
        app_Group = []
        
        print(g(' ►► Working with [') + m(rule_name) + g(']....'))
        all_policies = get_configured_policy(file_content=file_content)
        for each_rule in all_policies:
            if rule_name.lower() == each_rule[3].strip().lower():
                print_rule(each_rule,rule_name,Sources,Destinations,Services)
                user_input = input(' ►► Press [' + yellow('c') + '] to continue or any key to skip : ').lower()
                if user_input != 'c':
                    skipped_rules.append(each_line)
                    break
        else:
            Source_Zones = return_zones(Sources,dev=dev,file_content=file_content)#[[address,zone,[command]],[False,False,False]]   # command may == ''  
            obj1 = commands_zonesGroup(Source_Zones,Source_Zones_Group)#[[['command'],''],[['zone',[address,address']]]]
            for y in obj1[0]:#[['command'],'']
                for x in y:
                    if x != '':
                        final_commands.append(x)
            Source_Zones_Group = obj1[1] #[[zone1,[add1,add2]],[zone2,[add3,add4]]]

            Destination_Zones = return_zones(Destinations,dev=dev,file_content=file_content)#[[address,zone,[command]],[False,False,False]]   # command may == ''  
            obj4 = commands_zonesGroup(Destination_Zones,Destination_Zones_Group)#[[['command'],''],[['zone',[address,address']]]]
            for y in obj4[0]:
                for x in y:
                    if x != '':
                        final_commands.append(x)
            Destination_Zones_Group = obj4[1] #[[zone1,[add1,add2]],[zone2,[add3,add4]]]
            
            applications = return_application(Services,dev=dev,file_content=file_content) #[['app',['command1','command2']],[False,False]]  # command may == '' 
            obj2 = commands_appGroup(applications)#[ ,['app1','app2','app3']]
            for x in obj2[0]:
                if x != '':
                    final_commands.append(x)
            app_Group = obj2[1] #['app1','app2','app3'],

            obj3 = policy_maker(rule_name,Source_Zones_Group,Destination_Zones_Group,app_Group,dev=dev ,file_content=file_content)
            for x in obj3:
                if x != '':
                    final_commands.append(x)
               
            

    return [final_commands,policy_description]

def valid_ports(port):

    # Split the port by hyphen and check the parts
    parts = port.split('-')
    if len(parts) == 1 and parts[0].isdigit():
        return True
        
    elif len(parts) == 2 and parts[0].isdigit() and parts[1].isdigit():
        return True
    else:
        return False

def network_ip(ip_str,invalid_line):
    
    try:
    
        ip_network = ipaddress.IPv4Network(ip_str, strict=False)
        
        # Extract the IPv4 address without the subnet
        ipv4_address = ip_network.network_address
        return ipv4_address
        
    except Exception as e:
        wrong_inputs = f"Error: Invalid characters in the IPv4 address: \b{invalid_line} no action had taken."
        return ['error',wrong_inputs]
        
        
def ip_without_prefix(ip_str):
        if ip_str.find('/') != -1 :
            ipv4_prefix = ip_str.split('/')[0]
            return ipv4_prefix
            
      
def is_ip_in_range(ip, ip_prefix):
    try:
        # Parse the IP address and prefix into IPv4Network objects
        network = ipaddress.IPv4Network(ip_prefix, strict=False)
        ip_address = ipaddress.IPv4Address(ip)

        # Check if the IP address is within the range of the IP prefix
        if ip_address in network:
            return True
        else:
            return False
    except (ipaddress.AddressValueError, ValueError):
        return False
        
def most_specific_range(ip, ip_ranges):
    most_specific_match = None
    longest_prefix_length = -1

    ip = ipaddress.IPv4Address(ip)

    for ip_range in ip_ranges:
        try:
            network = ipaddress.IPv4Network(ip_range, strict=False)
            if ip in network and network.prefixlen > longest_prefix_length:
                most_specific_match = ip_range  # Store the actual IP range
                longest_prefix_length = network.prefixlen
        except (ipaddress.AddressValueError, ipaddress.NetmaskValueError):
            pass

    return most_specific_match


    
def is_valid_ipv4(ip):
    try:
        ipaddress.IPv4Address(ip)
        return True
    except ipaddress.AddressValueError:
        return False
        
def values_after(mylist,after_value):
    # Initialize variables
    found_Item = False
    found_values_between = []
    search_range = []
    if after_value == 'Src': search_range = ['Src. Address','Dst. Address']
    if after_value == 'Dst': search_range = ['Dst. Address','Ser.']
    if after_value == 'Ser': search_range = ['Ser.','Src. Address']
    
    
    # Iterate through the list
    for item in mylist:
        if item == search_range[0]:
            found_Item = True
        elif item == search_range[1]:
            found_Item = False
        elif found_Item and item != '':
            found_values_between.append(item)

    return found_values_between
def m(word):
    styled_text = colored(word, "magenta")
    return styled_text
def c(word):
    styled_text = colored(word, "cyan")
    return styled_text
def b(word):
    styled_text = colored(word, "blue")
    return styled_text
def yellow(word):
    styled_text = colored(word, "yellow")
    return styled_text
def g(word):
    styled_text = colored(word, "green")
    return styled_text


import os
import pandas as pd
from datetime import datetime
from srx_classes import devices
def extract_excel(sheet_data, file_name):
    # Get the current date and time
    current_datetime = datetime.now()
    formatted_datetime = current_datetime.strftime("[%d - %m - %Y]-[%H - %M]")  # Format the date and time as a string


    # Get the path to the "excel_output" folder
    output_directory = os.path.join("C:", devices.package_name, "Excel Outputs")

    # Construct the output file name with date and time
    output_file = os.path.join(output_directory, f'{file_name}-{formatted_datetime}.xlsx')

    # Create an ExcelWriter object with XlsxWriter as the engine
    writer = pd.ExcelWriter(output_file, engine='xlsxwriter')

    # Get the xlsxwriter workbook object
    workbook = writer.book
    
    for sheet_data_item in sheet_data:
        sheet_name = sheet_data_item['sheet_name']
        data = sheet_data_item['data']
        columns = sheet_data_item['columns']

        # Create a DataFrame for the current sheet
        df = pd.DataFrame(data, columns=columns)

        # Save the DataFrame in the current sheet
        df.to_excel(writer, sheet_name=sheet_name, index=False)

        # Get the worksheet object for the current sheet
        worksheet = writer.sheets[sheet_name]
        # Apply autofilter for the first row
        worksheet.autofilter(0, 0, 0, len(columns) - 1)
        

        # Autofit column widths for all columns
        for i, col in enumerate(columns):
            column_len = max(df[col].astype(str).apply(len).max(), len(col))
            worksheet.set_column(i, i, column_len + 5)  # Add a little extra space


        
        # Apply center alignment, vertical center alignment, and text wrapping to all cells
        center_alignment_wrap = workbook.add_format({'align': 'center', 'valign': 'vcenter', 'text_wrap': True})

        for row in range(df.shape[0]):
            for col in range(df.shape[1]):
                worksheet.write(row + 1, col, df.iat[row, col], center_alignment_wrap)

        # Freeze the first row
        worksheet.freeze_panes(1, 0)

    # Close the Pandas ExcelWriter, which will save the Excel file
    writer.close()

    print(f'Data saved to {output_file}')





from jnpr.junos.exception import ConfigLoadError
from jnpr.junos.exception import CommitError


     
def load_configurations(cu,config_set):
    try:    
        command_err=''
        #load configuration changes
        print (">Loading configuration in progress...")
        
        for command in config_set.splitlines():
            command_err = command
            cu.load(command, format='set')
            
        print('********************* The Changes in Configuration*********************')
        #print the candidate configuration on screen
        cu.pdiff()
        
        print (">Configuration loaded succesfully\n")
        print("Press any key to continue...")
        input()
    except (ConfigLoadError, CommitError) as err:
        print ('Error at: ' ,command_err)
        print ("Unable to load configuration changes:\n{0}".format(err))
        print("Press any key to continue...")
        input()
        
def load_list_set_configurations(cu,config_set):
    try:    
        command_err=''
        #load configuration changes
        
        print (">Loading configuration in progress...")
        
        for command in config_set:
            command_err = command
            cu.load(command, format='set')
            
        print('********************* The Changes in Configuration*********************')
        
        #print the candidate configuration on screen
        cu.pdiff()
        
        print (">Configuration loaded succesfully\n")
        print("Press any key to continue...")
        input()
    except (ConfigLoadError, CommitError) as err:
        print ('Error at: ' ,command_err)
        print ("Unable to load configuration changes:\n{0}".format(err))
        print("Press any key to continue...")
        input()
        


from setuptools import setup, find_packages

setup(
    name='psrx',
    version='1.3',
    packages=find_packages(),
     entry_points={
        'console_scripts': [
            'psrx = srx_classes.main_script:main',
        ],
    },
    install_requires=[
        '2to3==1.0',
        'altgraph==0.17.4',
        'argcomplete==3.1.2',
        'bcrypt==4.0.1',
        'cffi==1.16.0',
        'click==8.1.7',
        'clipboard==0.0.4',
        'colorama==0.4.6',
        'cryptography==41.0.4',
        'ecdsa==0.18.0',
        'et-xmlfile==1.1.0',
        'gmpy2==2.1.5',
        'Jinja2==3.1.2',
        'junos-eznc==2.6.8',
        'lxml==4.9.3',
        'markdown-it-py==3.0.0',
        'MarkupSafe==2.1.3',
        'mdurl==0.1.2',
        'ncclient==0.6.13',
        'netaddr==0.9.0',
        'numpy==1.26.1',
        'openpyxl==3.1.2',
        'packaging==23.2',
        'pandas==2.1.1',
        'paramiko==3.3.1',
        'pefile==2023.2.7',
        'progressbar==2.5',
        'prompt-toolkit==3.0.21',
        'pycparser==2.21',
        'Pygments==2.16.1',
        'pyinstaller==6.0.0',
        'pyinstaller-hooks-contrib==2023.9',
        'PyNaCl==1.5.0',
        'pyparsing==3.1.1',
        'pyperclip==1.8.2',
        'pyreadline==2.1',
        'pyserial==3.5',
        'python-dateutil==2.8.2',
        'pytz==2023.3.post1',
        'pywin32-ctypes==0.2.2',
        'PyYAML==6.0.1',
        'regex==2023.10.3',
        'rich==13.6.0',
        'scp==0.14.5',
        'shellingham==1.5.3',
        'six==1.16.0',
        'termcolor==2.3.0',
        'tqdm==4.66.1',
        'transitions==0.9.0',
        'typer==0.9.0',
        'typing_extensions==4.8.0',
        'tzdata==2023.3',
        'wcwidth==0.2.8',
        'XlsxWriter==3.1.9',
        'yamlordereddictloader==0.4.2',
    ],
    author='Eslam',
    description='doing all configuration by one click',
    long_description=open('README.md').read(),
    license='MIT',
)



from srx_classes import _configurations_
from termcolor import colored
application_description = 'Application Pushed through Netconf'





excelude_applications = [
    'application_group',
    'application-group',
    'application group',
    ''
    ]
    
def application_applications_set(dev, input_any_applications_to_configure, file_content = None): #dev, input_any_applications_to_configure

    invalid_application_set = []
    
    valid_application_count = 0 
    
    commands=[]
    
    already_excited=[]
    
    duplicated_in_commands = []
    lines = input_any_applications_to_configure.strip().split('\n')
    
     # [[each line.split], [each line.split], [each line.split]]
    result = [line.split('\t') for line in lines]

    for each_line1 in result:
    
# if >>  For application only >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>.
        if len(each_line1) == 6 and each_line1[5] != '':
        
            # if desctription has space between it's word replace it with _
            each_line = [element.strip().replace('\t','').replace('\xa0','').replace(' ', '_').replace('&', 'And').replace('(','').replace(')','').replace('\r','') for element in each_line1]

 #==================== For Application Only ==================================================================================================
            
            valid_Src_Port = valid_ports(each_line[4])
            valid_Dst_Port = valid_ports(each_line[5])
            if dev != None:
                retuen_match_in_list = _configurations_.retrive_configurations('application' , each_line[1], dev=dev)
            if file_content != None:
                retuen_match_in_list = _configurations_.retrive_configurations('application' , each_line[1], file_content=file_content)
                    
            if each_line[2] == '' and valid_Dst_Port: # ['Application', name, '', prtocol , any, port ]
                if not(valid_Src_Port): # if source port not exist
                    if not(f'set applications application {each_line[1]} protocol {each_line[3].lower()}' in commands):
                        if len(retuen_match_in_list) == 0:    
                            commands.append(f'set applications application {m(each_line[1])} protocol {yellow(each_line[3].lower())}')
                            commands.append(f'set applications application {m(each_line[1])} destination-port {yellow(each_line[5])}')
                            valid_application_count +=1
                            continue
                        else:
                            already_excited.append(f'The application [{m(each_line[1])}] Already excited. No action had taken.')
                            continue
                    else:
                        duplicated_in_commands.append(f'The application [{m(each_line[1])}] Is duplicated and done once. No action had taken.')
                        continue
                
                if valid_Src_Port: # if source port exist  
                    if not(f'set applications application {each_line[1]} protocol {each_line[3].lower()}' in commands):
                        if len(retuen_match_in_list) == 0:    
                            commands.append(f'set applications application {m(each_line[1])} protocol {yellow(each_line[3].lower())}')
                            commands.append(f'set applications application {m(each_line[1])} source-port {yellow(each_line[4])}')
                            commands.append(f'set applications application {m(each_line[1])} destination-port {yellow(each_line[5])}')
                            valid_application_count +=1
                            continue
                        else:
                            already_excited.append(f'The application [{m(each_line[1])}] Already excited. No action had taken.')
                            continue
                    else:
                        duplicated_in_commands.append(f'The application [{m(each_line[1])}] Is duplicated and done once. No action had taken.')
                        continue
            elif each_line[2] == '' and not(valid_Dst_Port):
                
                invalid_application_set.append(f' You enter invalid destination-port [{yellow(each_line[5])}] in the application name[{m(each_line[1])}]. no action had taken.')
                continue   
               
           
            if each_line[2] != '' and valid_Dst_Port: # ['Application', name, description, prtocol , any, port ] 
                if not(valid_Src_Port): # if source port not exist
                    if not(f'set applications application {each_line[1]} description {each_line[2]} protocol {each_line[3].lower()}' in commands):
                        if len(retuen_match_in_list) == 0:    
                            commands.append(f'set applications application {m(each_line[1])} description {yellow(each_line[2])} protocol {yellow(each_line[3].lower())}')
                            commands.append(f'set applications application {m(each_line[1])} description {yellow(each_line[2])} destination-port {yellow(each_line[5])}')
                            valid_application_count +=1
                            continue
                        else:
                            already_excited.append(f'The application [{m(each_line[1])}] Already excited. No action had taken.')
                            continue
                    else:
                        duplicated_in_commands.append(f'The application [{m(each_line[1])}] Is duplicated and done once. No action had taken.')
                        continue
                    
                if valid_Src_Port: # if source port exist
                    if not(f'set applications application {each_line[1]} description {each_line[2]} protocol {each_line[3].lower()}' in commands):
                        if len(retuen_match_in_list) == 0:    
                            commands.append(f'set applications application {m(each_line[1])} description {yellow(each_line[2])} protocol {yellow(each_line[3].lower())}')
                            commands.append(f'set applications application {m(each_line[1])} description {yellow(each_line[2])} source-port {yellow(each_line[4])}')
                            commands.append(f'set applications application {m(each_line[1])} description {yellow(each_line[2])} destination-port {yellow(each_line[5])}')
                            valid_application_count +=1
                            continue
                        else:
                            already_excited.append(f'The application [{m(each_line[1])} ] Already excited. No action had taken.')
                            continue
                    else:
                        duplicated_in_commands.append(f'The application [{m(each_line[1])}] Is duplicated and done once. No action had taken.')
                        continue
            elif each_line[2] != '' and not(valid_Dst_Port):
                invalid_application_set.append(f' You enter invalid destination-port [{yellow(each_line[5])}] in the application name[{m(each_line[1])}]. no action had taken.')
                continue          
            
                    
  # if >>  For application group >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>.  
  
        #inilaize each_line1 add more elements '' if have less than 4
        if len(each_line1) < 6:
            for more_element in range(6 - len(each_line1)):
                each_line1.append('')
                
        each_line = [element.strip().replace('\t','').replace('\xa0','').replace(' ', '_').replace('&', 'And').replace('(','').replace(')','').replace('\r','') for element in each_line1]
        
        if len(each_line) == 6 and each_line[5] == '':
            
            # send column 2 value to retrived data > return true or false
            if dev != None:
                is_excist_application = _configurations_.retrive_configurations('application' , each_line[1], is_exist= 'yes', dev = dev)
                retuen_match_in_list = _configurations_.retrive_configurations('application-set' , [each_line[0],each_line[1]], is_exist= 'yes', dev=dev)
            if file_content != None:
                
                is_excist_application = _configurations_.retrive_configurations('application' , each_line[1], is_exist= 'yes', file_content=file_content)
                
                retuen_match_in_list = _configurations_.retrive_configurations('application-set' , [each_line[0],each_line[1]], file_content=file_content)
                #retuen_match_in_list return a list that contain [logical app match , normal match] or [normal match ] or []
            if is_excist_application and not(each_line[0].lower() in [s.lower() for s in excelude_applications]):   # ['address-set', address, '', ''] 

                if not(each_line[0].lower() in [s.lower() for s in excelude_applications]):   # ['address-set', address, '', ''] 

                    if not(f'set applications application-set {each_line[0]} application {each_line[1]}' in commands):
                    
                        if len(retuen_match_in_list) == 0:

                            commands.append(f'set applications application-set {c(each_line[0])} application {m(each_line[1])}')
                            valid_application_count +=1
                            continue
                            
                        else:
                        
                            already_excited.append(f'The application [{m(each_line[1])}] Already excited in application-set called [ {c(each_line[0])} ]. No action had taken.')
                            continue
                            
                    else:
                    
                        duplicated_in_commands.append(f'The application [{m(each_line[1])}] Is duplicated and done once. No action had taken.')
                        continue
            else:
            
                invalid_application_set.append(f'The application [{m(each_line[1])}] Not exist. So you can not add it to application-set [ {c(each_line[0])} ]. No action had taken.')
                continue
                
                
    return [commands, invalid_application_set, application_description,valid_application_count,duplicated_in_commands,already_excited]

def valid_ports(port):

    # Split the port by hyphen and check the parts
    parts = port.split('-')
    if len(parts) == 1 and parts[0].isdigit():
        return True
        
    elif len(parts) == 2 and parts[0].isdigit() and parts[1].isdigit():
        return True
    else:
        return False

def m(word):
    styled_text = colored(word, "magenta", attrs=["bold"])
    return styled_text
def c(word):
    styled_text = colored(word, "cyan", attrs=["bold"])
    return styled_text
def b(word):
    styled_text = colored(word, "blue", attrs=["bold"])
    return styled_text
def yellow(word):
    styled_text = colored(word, "yellow", attrs=["bold"])
    return styled_text
def g(word):
    styled_text = colored(word, "green", attrs=["bold"])
    return styled_text